use std::{collections::HashMap, net::UdpSocket, time::SystemTime};

use bevy::{
    diagnostic::{FrameTimeDiagnosticsPlugin, LogDiagnosticsPlugin},
    prelude::*,
};

use bevy_egui::{EguiContext, EguiPlugin};
use bevy_rapier2d::prelude::*;

use bevy_renet::{
    renet::{RenetServer, ServerAuthentication, ServerConfig, ServerEvent},
    RenetServerPlugin,
};

use daggmask_shared::{
    server_connection_config, setup_level, spawn_projectile, ClientChannel, Jumper, NetworkFrame,
    Player, PlayerCommand, PlayerInput, Projectile, ServerChannel, ServerMessages, PROTOCOL_ID,
};

use renet_visualizer::RenetServerVisualizer;

#[derive(Debug, Default)]
pub struct ServerLobby {
    pub players: HashMap<u64, Entity>,
}

#[derive(Debug, Default)]
struct NetworkTick(u32);

// Clients last received ticks
#[derive(Debug, Default)]
struct ClientTicks(HashMap<u64, Option<u32>>);

const PLAYER_MOVE_SPEED: f32 = 50.0;

fn new_renet_server() -> RenetServer {
    let server_addr = "127.0.0.1:5000".parse().unwrap();
    let socket = UdpSocket::bind(server_addr).unwrap();
    let connection_config = server_connection_config();
    let server_config =
        ServerConfig::new(64, PROTOCOL_ID, server_addr, ServerAuthentication::Unsecure);
    let current_time = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .unwrap();
    RenetServer::new(current_time, server_config, connection_config, socket).unwrap()
}

fn main() {
    let mut app = App::new();
    app.add_plugins(DefaultPlugins);

    app.add_plugin(RenetServerPlugin);
    app.add_plugin(RapierPhysicsPlugin::<NoUserData>::default());
    app.add_plugin(RapierDebugRenderPlugin::default());
    app.add_plugin(FrameTimeDiagnosticsPlugin::default());
    app.add_plugin(LogDiagnosticsPlugin::default());
    app.add_plugin(EguiPlugin);

    app.insert_resource(ServerLobby::default());
    app.insert_resource(NetworkTick(0));
    app.insert_resource(ClientTicks::default());
    app.insert_resource(new_renet_server());
    app.insert_resource(RenetServerVisualizer::<200>::default());

    app.add_system_to_stage(CoreStage::Update, server_update_system);
    app.add_system_to_stage(CoreStage::PreUpdate, server_network_sync);
    app.add_system(move_players_system);
    app.add_system(jump_system);
    app.add_system(jump_reset);
    app.add_system(update_projectiles_system);
    app.add_system(update_visualizer_system);
    app.add_system(despawn_projectile_system);
    app.add_system_to_stage(CoreStage::PostUpdate, projectile_on_removal_system);

    app.add_startup_system(setup_level);
    app.add_startup_system(setup_simple_camera);

    app.add_startup_stage("floor_setup", SystemStage::single(spawn_floor));

    app.run();
}

#[allow(clippy::too_many_arguments)]
fn server_update_system(
    mut server_events: EventReader<ServerEvent>,
    pos_query: Query<(&GlobalTransform, &Player)>,
    mut commands: Commands,
    mut lobby: ResMut<ServerLobby>,
    mut server: ResMut<RenetServer>,
    mut visualizer: ResMut<RenetServerVisualizer<200>>,
    mut client_ticks: ResMut<ClientTicks>,
    players: Query<(Entity, &Player)>,
) {
    for event in server_events.iter() {
        match event {
            ServerEvent::ClientConnected(id, _) => {
                println!("Player {} connected.", id);

                visualizer.add_client(*id);

                // Initialize other players for this new client
                for (entity, player) in players.iter() {
                    let message = bincode::serialize(&ServerMessages::PlayerCreate {
                        id: player.id,
                        entity,
                    })
                    .unwrap();
                    server.send_message(*id, ServerChannel::ServerMessages.id(), message);
                }

                // Spawn new player
                let player_entity = commands
                    .spawn_bundle(SpriteBundle {
                        sprite: Sprite {
                            color: Color::rgb(0., 0.47, 1.),
                            custom_size: Some(Vec2::new(5., 5.)),
                            ..Default::default()
                        },
                        ..Default::default()
                    })
                    .insert_bundle(TransformBundle::from(Transform::from_xyz(0., 0., 0.)))
                    .insert(RigidBody::Dynamic)
                    .insert(LockedAxes::ROTATION_LOCKED)
                    .insert(Collider::cuboid(5., 5.))
                    .insert(ActiveEvents::COLLISION_EVENTS)
                    .insert(PlayerInput::default())
                    .insert(Velocity::default())
                    .insert(Player { id: *id })
                    .insert(Jumper {
                        jump_impulse: 80.,
                        is_jumping: false,
                    })
                    .id();

                lobby.players.insert(*id, player_entity);

                let message = bincode::serialize(&ServerMessages::PlayerCreate {
                    id: *id,
                    entity: player_entity,
                })
                .unwrap();

                server.broadcast_message(ServerChannel::ServerMessages.id(), message);
            }
            ServerEvent::ClientDisconnected(id) => {
                println!("Player {} disconnected.", id);

                visualizer.remove_client(*id);
                client_ticks.0.remove(id);

                if let Some(player_entity) = lobby.players.remove(id) {
                    commands.entity(player_entity).despawn();
                }

                let message =
                    bincode::serialize(&ServerMessages::PlayerRemove { id: *id }).unwrap();

                server.broadcast_message(ServerChannel::ServerMessages.id(), message);
            }
        }
    }

    for client_id in server.clients_id().into_iter() {
        while let Some(message) = server.receive_message(client_id, ClientChannel::Command.id()) {
            let command: PlayerCommand = bincode::deserialize(&message).unwrap();

            match command {
                PlayerCommand::BasicAttack { direction } => {
                    for (transform, player) in pos_query.iter() {
                        if player.id == client_id {
                            let origin =
                                Vec2::new(transform.translation().x, transform.translation().y);

                            info!(
                                "spawned attack on server at origin {} with direction {}",
                                origin, direction
                            );

                            let entity = spawn_projectile(&mut commands, origin, direction);
                            let message = ServerMessages::SpawnProjectile { entity, origin };
                            let message = bincode::serialize(&message).unwrap();

                            server.broadcast_message(ServerChannel::ServerMessages.id(), message);
                        }
                    }
                }
            }
        }

        while let Some(message) = server.receive_message(client_id, ClientChannel::Input.id()) {
            let input: PlayerInput = bincode::deserialize(&message).unwrap();

            client_ticks.0.insert(client_id, input.most_recent_tick);

            if let Some(player_entity) = lobby.players.get(&client_id) {
                commands.entity(*player_entity).insert(input);
            }
        }
    }
}

fn update_projectiles_system(
    mut commands: Commands,
    mut projectiles: Query<(Entity, &mut Projectile)>,
    time: Res<Time>,
) {
    for (entity, mut projectile) in projectiles.iter_mut() {
        projectile.duration.tick(time.delta());

        if projectile.duration.finished() {
            commands.entity(entity).despawn();
        }
    }
}

fn update_visualizer_system(
    mut egui_context: ResMut<EguiContext>,
    mut visualizer: ResMut<RenetServerVisualizer<200>>,
    server: Res<RenetServer>,
) {
    visualizer.update(&server);
    visualizer.show_window(egui_context.ctx_mut());
}

#[allow(clippy::type_complexity)]
fn server_network_sync(
    mut tick: ResMut<NetworkTick>,
    mut server: ResMut<RenetServer>,
    networked_entities: Query<(Entity, &GlobalTransform), Or<(With<Player>, With<Projectile>)>>,
) {
    let mut frame = NetworkFrame::default();

    for (entity, transform) in networked_entities.iter() {
        frame.entities.entities.push(entity);
        frame
            .entities
            .translations
            .push(transform.translation().into());
    }

    frame.tick = tick.0;
    tick.0 += 1;

    let sync_message = bincode::serialize(&frame).unwrap();

    server.broadcast_message(ServerChannel::NetworkFrame.id(), sync_message);
}

fn move_players_system(mut query: Query<(&mut Velocity, &PlayerInput)>) {
    for (mut velocity, input) in query.iter_mut() {
        let x_axis = -(input.left as i8) + input.right as i8;

        let mut move_delta = Vec2::new(x_axis as f32, 0.);
        if move_delta != Vec2::ZERO {
            move_delta /= move_delta.length();
        }

        velocity.linvel.x = move_delta.x * PLAYER_MOVE_SPEED;
    }
}

fn jump_system(mut players: Query<(&mut Jumper, &mut Velocity, &PlayerInput)>) {
    for (mut jumper, mut velocity, input) in players.iter_mut() {
        if input.space && !jumper.is_jumping {
            velocity.linvel.y = jumper.jump_impulse;
            jumper.is_jumping = true;
        }
    }
}

fn jump_reset(
    mut query: Query<(Entity, &mut Jumper), Without<Projectile>>,
    mut collision_events: EventReader<CollisionEvent>,
) {
    for collision_event in collision_events.iter() {
        for (entity, mut jumper) in query.iter_mut() {
            set_jumping_false_if_touching_floor(entity, &mut jumper, collision_event);
        }
    }
}

fn set_jumping_false_if_touching_floor(
    entity: Entity,
    jumper: &mut Jumper,
    event: &CollisionEvent,
) {
    if let CollisionEvent::Started(h1, h2, _) = event {
        if h1.id() == entity.id() || h2.id() == entity.id() {
            jumper.is_jumping = false
        }
    }
}

fn setup_simple_camera(mut commands: Commands) {
    commands.spawn_bundle(Camera2dBundle::default());
}

fn despawn_projectile_system(
    mut commands: Commands,
    mut collision_events: EventReader<CollisionEvent>,
    projectile_query: Query<Option<&Projectile>>,
) {
    for collision_event in collision_events.iter() {
        if let CollisionEvent::Started(entity1, entity2, _) = collision_event {
            if let Ok(Some(_)) = projectile_query.get(*entity1) {
                commands.entity(*entity1).despawn();
            }

            if let Ok(Some(_)) = projectile_query.get(*entity2) {
                commands.entity(*entity2).despawn();
            }
        }
    }
}

fn projectile_on_removal_system(
    mut server: ResMut<RenetServer>,
    removed_projectiles: RemovedComponents<Projectile>,
) {
    for entity in removed_projectiles.iter() {
        let message = ServerMessages::DespawnProjectile { entity };
        let message = bincode::serialize(&message).unwrap();

        server.broadcast_message(ServerChannel::ServerMessages.id(), message);
    }
}

fn spawn_floor(mut commands: Commands) {
    let width = 300.;
    let height = 3.;

    commands
        .spawn_bundle(SpriteBundle {
            sprite: Sprite {
                color: Color::rgb(0., 0.47, 1.),
                custom_size: Some(Vec2::new(width, height)),
                ..Default::default()
            },
            ..Default::default()
        })
        .insert_bundle(TransformBundle::from(Transform::from_xyz(-50., 0., 0.)))
        .insert(RigidBody::Fixed)
        .insert(Collider::cuboid(width / 2., height / 2.))
        .insert(ActiveEvents::COLLISION_EVENTS);
}
